package bst;

public class BinarySearchTree<E> implements IBinarySearchTree<E>{

    private BSTNode<E> root;
    private int size;

    public BinarySearchTree() {
        root = null;
        size = 0;
    }

    public BinarySearchTree(BSTNode<E> root) { //  for testing purposes
        this.root = root;
    }

    @Override
    public boolean contains(Object o) {
        Comparable<E> searchKey = (Comparable<E>) o; // throws ClassCastException if Object is not comparable to E
        return containsHelper(searchKey, root);
    }

    private boolean containsHelper(Comparable<E> e, BSTNode<E> treeRoot) {
        if (treeRoot == null) {
            return false;
        }
        int comparison = e.compareTo(treeRoot.getValue());
        boolean isLeaf = treeRoot.isLeaf();
        if (comparison == 0) {
            return true;
        } else if (comparison < 0) {
            // if this is a leaf, STOP! if this is internal, then continue the search
            // to the left subtree
            return !isLeaf && containsHelper(e, ((BSTInternal<E>) treeRoot).getLeft());
        } else {
            // if this is a leaf, STOP! if this is internal, then continue the search
            // to the right subtree
            return !isLeaf && containsHelper(e, ((BSTInternal<E>) treeRoot).getRight());
        }
    }

    @Override
    public boolean add(E e) {
        if (root == null) {
            root = new BSTLeaf<>(e);
            size++;
            return true;
        } else {
            BSTNode<E> theTree = addHelper(e, root);
            if (theTree == null) {
                return false;
            } else {
                root = theTree;
                size++;
                return true;
            }
        }
    }

    private BSTNode<E> addHelper(E e, BSTNode<E> root) {
        // Java's bst.TreeSet is implemented with a self-balancing BST, and
        // Sets don't store duplicates, so we will reject duplicates in this implementation.
        // This isn't necessarily a requirement of BSTs, but we're doing it here.
        int comparison = ((Comparable<E>) e).compareTo(root.getValue());
        if (comparison == 0) {
            return null; // no changes made
        }
        if (comparison < 0) {
            if (root.isLeaf()) {
                return new BSTInternal<>(root.getValue(), new BSTLeaf<>(e), null);
            } else {
                BSTInternal<E> asInternal = (BSTInternal<E>) root;
                if (asInternal.getLeft() == null) {
                    asInternal.setLeft(new BSTLeaf<E>(e));
                    return root;
                }
                asInternal.setLeft(addHelper(e, asInternal.getLeft()));
                return asInternal;
            }
        } else {
            if (root.isLeaf()) {
                return new BSTInternal<>(root.getValue(), null, new BSTLeaf<>(e));
            } else {
                BSTInternal<E> asInternal = (BSTInternal<E>) root;
                if (asInternal.getRight() == null) {
                    asInternal.setRight(new BSTLeaf<E>(e));
                    return root;
                }
                asInternal.setRight(addHelper(e, asInternal.getRight()));
                return asInternal;
            }
        }
    }

    @Override
    public boolean remove(Object o) {
        BSTInternal<E> parent = null;
        BSTNode<E> current = root;
        Comparable<E> searchKey = (Comparable<E>) o; // ClassCastException possible
        while (current != null) {
            int comparison = searchKey.compareTo(current.getValue());
            if (comparison == 0) { // found it!
                if (current.isLeaf()) { // CASE 1: Removing a leaf
                    if (parent == null) {
                        root = null;
                    } else if (parent.getLeft() == current) {
                        parent.setLeft(null);
                    } else {
                        parent.setRight(null);
                    }
                    size--;
                    return true;
                }
                BSTInternal<E> currAsInternal = (BSTInternal<E>) current;
                if (currAsInternal.getRight() == null) { // CASE 2: Removing node with only left child
                    if (parent == null) {
                        root = currAsInternal.getLeft();
                    } else if (parent.getLeft() == currAsInternal) {
                        parent.setLeft(currAsInternal.getLeft());
                    } else {
                        parent.setRight(currAsInternal.getLeft());
                    }
                } else if (currAsInternal.getLeft() == null) { // CASE 3: Removing node with only right child
                    if (parent == null) {
                        root = currAsInternal.getRight();
                    } else if (parent.getLeft() == currAsInternal) {
                        parent.setLeft(currAsInternal.getRight());
                    } else {
                        parent.setRight(currAsInternal.getRight());
                    }
                } else { // CASE 4: Removing node with both children
                    BSTNode<E> successor = currAsInternal.getRight();
                    while (!successor.isLeaf() && ((BSTInternal<E>) successor).getLeft() != null) { // find successor node of current
                        successor = ((BSTInternal<E>) successor).getLeft();
                    }
                    E data = successor.getValue();
                    remove(data); // remove the successor node
                    current.setValue(data); // replace current node's data with its successor
                }
                size--;
                return true;

            } else if (comparison < 0) {
                if (current.isLeaf()) {
                    return false;
                }
                parent = ((BSTInternal<E>) current);
                current = ((BSTInternal<E>) current).getLeft();
            } else {
                if (current.isLeaf()) {
                    return false;
                }
                parent = ((BSTInternal<E>) current);
                current = ((BSTInternal<E>) current).getRight();

            }
        }
        return false;
    }

    public BSTNode<E> root() {
        return this.root;
    }

    public void inOrderTraverse() {
        inOrderTraverseHelper(root);
    }

    public void inOrderTraverseHelper(BSTNode<E> treeRoot) {
        // BASE CASE(s)
        if (treeRoot.isLeaf()) {
            System.out.println(treeRoot.getValue());
            return;
        }

        // GET LEFT & RIGHT SUBTREES
        BSTNode<E> leftSubtree = ((BSTInternal<E>) treeRoot).getLeft();
        BSTNode<E> rightSubtree = ((BSTInternal<E>) treeRoot).getRight();

        // VISIT LEFT SUBTREE
        if (leftSubtree != null) {
            inOrderTraverseHelper(leftSubtree);
        }
        // VISIT NODE
        System.out.println(treeRoot.getValue());

        // VISIT RIGHT SUBTREE
        if (rightSubtree != null) {
            inOrderTraverseHelper(rightSubtree);
        }
    }

    public static void main(String[] args) {
        BinarySearchTree<Integer> removalTree = new BinarySearchTree<>();
        Integer[] insertions = {19, 14, 25, 21, 30, 27, 68 };
        for (int insertion : insertions) {
            removalTree.add(insertion);
        }

        removalTree.inOrderTraverse();
    }







    public E find(Object o) {
        Comparable<E> searchKey = (Comparable<E>) o; // throws ClassCastException if Object is not comparable to E
        return findHelper(searchKey, root);
    }

    private E findHelper(Comparable<E> e, BSTNode<E> treeRoot) {
        if (treeRoot == null) {
            return null;
        }
        int comparison = e.compareTo(treeRoot.getValue());
        boolean isLeaf = treeRoot.isLeaf();
        if (comparison == 0) {
            return treeRoot.getValue();
        } else if (comparison < 0) {
            if (isLeaf) {
                return null;
            }
            // if this is a leaf, STOP! if this is internal, then continue the search
            // to the left subtree
            return findHelper(e, ((BSTInternal<E>) treeRoot).getLeft());
        } else {
            if (isLeaf) {
                return null;
            }
            // if this is a leaf, STOP! if this is internal, then continue the search
            // to the right subtree
            return findHelper(e, ((BSTInternal<E>) treeRoot).getRight());
        }
    }


}
